from ..service import ReceivingService, ThreadService
from ..wire import FloatWire

class SaveRaw(ThreadService, ReceivingService):
    input_wire_class=FloatWire
    def __init__(self, filename=None):
        super().__init__()
        
        if filename is None:
            filename="data.raw"
        
        self.outputfile=filename
    
    def start_process(self):    
        with open(self.outputfile, 'wb') as self.output:
            super().start_process()
        self.print_message(f"done writing to {self.outputfile}")

    def process(self, data):
        self.output.write(data["data"].tobytes())



from ..service import ReceivingService, ThreadService
from ..wire import Int16Wire

class SaveWav(ThreadService, ReceivingService):
    input_wire_class=Int16Wire
    
    def __init__(self, filename=None):
        super().__init__()

        if filename is None:
            filename="data.wav"
        
        self.outputfile=filename
        
    def start_process(self):
        global wave
        
        try:
          import wave
        except Exception as ex:
            self.print_message( f"import error: {str(ex)}")

        with wave.open(self.outputfile, 'wb') as self.output:
            self.output.setnchannels(1)
            self.output.setsampwidth(2)
            self.output.setframerate(self.input_wire.RATE)
            
            super().start_process()
        self.print_message(f"done writing to {self.outputfile}")

    def process(self, data):
        self.output.writeframes(data["data"].tobytes())
                



import configparser
import numpy

from datetime import datetime

def calibration_pair(s):
    """
      type for argument parser in order to interpret space seperated list of
      (value,energy) pairs.
    """
    return list(map(float, s.split(",")))

def get_calibration_coeff(data=None, order=2, no_offset=False, 
        return_coeff=False,write_ini=False, read_ini=False):
    """
      calculate calibration coefficients from (raw, energy) pairs
      
      Arguments:
        data: list of (raw_value, energy) pairs of calibration data
        order: order of polynomial fit order =1 or 2 is recommended
        no_offset: boolean indicating whether to force offset to be zero,
        return_coeff: boolean indicating  whether to return raw coefficients or 
                      offset, scale and drift (default).
        write_ini: whether to write out a ini file with calibration data
        read_ini: whether to read ini file with calibration, ignoring any other arguments
        
      Returns:
        tuple of offset, scale and drift: (E=offset+scale*d*(1+d*x))
          or 
        raw coefficients c_i: E=c_0 + c_1*d + ... + c_n*d**n
      
    """
    if read_ini:
        return read_calibration_file()
    
    if data is None or len(data)==0:
        raise Exception("no data, provide data ..")

    if no_offset:
        _offset=0
    else:
        _offset=1
    
    if order>len(data)-_offset:
        print("not enough data points, reducing order")
        order=len(data)-_offset
    
    data=numpy.array(data)
    x=data[:,0]
    y=data[:,1]
    
    a=numpy.zeros((len(x), order+_offset))
    
    for i in range(order+_offset):
        a[:,i]=x**(i+1-_offset)
    
    p=numpy.linalg.pinv(a)
    
    coeff=p.dot(y)
    
    if no_offset:
      coeff=numpy.concatenate([[0],coeff])
    
    offset=coeff[0]
    scale=coeff[1]
    if order>=2:
      drift=coeff[2]/coeff[1]
    
    # ~ print(f"raw coefficents: {coeff}")
    # ~ print(f"the offset is: {offset}")
    # ~ print(f"the scale is: {scale}")
    # ~ print(f"the drift is: {drift}")
    
    if write_ini:
        write_calibration_file(data,order=order,no_offset=no_offset)
    
    if return_coeff:
        return coeff
    else:
        return offset,scale,drift

def read_calibration_file(filename="gammalab.ini"):

    config = configparser.ConfigParser()
    config.read(filename)
    
    label=config["calibration"].get("label","(no label)")
    s=config["calibration"].get("data").split(" ")
    s=[x for x in s if x!=""]
    s=[calibration_pair(x) for x in s]
      
    order=config["calibration"].getint("order", 1)
    no_offset=config["calibration"].getboolean("no_offset", True)
    print(f"[util] calibration read from {filename}, labelled {label}")
    
    return get_calibration_coeff(s, order=order, no_offset=no_offset)

def write_calibration_file(data, filename="gammalab.ini", order=1, no_offset=True, label=None):
    data=" ".join([f"{x[0]},{x[1]}" for x in data])
    if label is None:
        label=f"{datetime.now()}"
    config = configparser.ConfigParser()
    config["calibration"]=dict(data=data, order=order, no_offset=no_offset, label=label)
    with open(filename,'w') as f:
        config.write(f)
    print(f"[util] calibration written to {filename}")



from ..service import ReceivingService, SourceService, ThreadService
from ..wire import RawWire, FloatWire, Int16Wire, NumpyWire

import numpy

class Identity(ThreadService, SourceService, ReceivingService):
    input_wire_class=RawWire
    output_wire_class=RawWire
    
    def output_protocol(self, wire):
        super().output_protocol(wire)
        wire.CHANNELS=self.input_wire.CHANNELS
        wire.RATE=self.input_wire.RATE
        wire.FORMAT=self.input_wire.FORMAT

    def process(self, data):
        return data

class Raw2Numpy(ThreadService, SourceService, ReceivingService):
    input_wire_class=RawWire
    output_wire_class=NumpyWire

    def output_protocol(self, wire):
        super().output_protocol(wire)
        wire.CHANNELS=self.input_wire.CHANNELS
        wire.RATE=self.input_wire.RATE
        wire.FORMAT=self.input_wire.FORMAT

    def process(self, data):
        data["data"]=numpy.frombuffer(data["data"],dtype=self.input_wire.FORMAT)
        return data

class Numpy2Raw(ThreadService, SourceService, ReceivingService):
    input_wire_class=NumpyWire
    output_wire_class=RawWire

    def output_protocol(self, wire):
        super().output_protocol(wire)
        wire.CHANNELS=self.input_wire.CHANNELS
        wire.RATE=self.input_wire.RATE
        wire.FORMAT=self.input_wire.FORMAT

    def process(self, data):
        data["data"]=data["data"].tobytes()
        return data


class Raw2Float(ThreadService, SourceService, ReceivingService):
    input_wire_class=RawWire
    output_wire_class=FloatWire

    def output_protocol(self, wire):
        super().output_protocol(wire)
        wire.CHANNELS=self.input_wire.CHANNELS
        wire.RATE=self.input_wire.RATE
        wire.FORMAT="float32"

    def process(self, data):
        if self.input_wire.FORMAT=="float32":
          data["data"]=numpy.frombuffer(data["data"],dtype=self.input_wire.FORMAT)
        elif self.input_wire.FORMAT=="int16":
          data["data"]=numpy.frombuffer(data["data"],dtype=self.input_wire.FORMAT)
          data["data"]=data["data"].astype("float32")/32768
        else:
          self.print_message("unknown data format in wire")
          return None
        return data

class Float2Raw(ThreadService, SourceService, ReceivingService):
    input_wire_class=FloatWire
    output_wire_class=RawWire

    def output_protocol(self, wire):
        super().output_protocol(wire)
        wire.CHANNELS=self.input_wire.CHANNELS
        wire.RATE=self.input_wire.RATE
        wire.FORMAT=self.input_wire.FORMAT

    def process(self, data):
        data["data"]=data["data"].tobytes()
        return data

class DownSampleMaxed(ThreadService, SourceService, ReceivingService):
    input_wire_class=FloatWire
    output_wire_class=FloatWire

    def __init__(self, factor=8):
        super().__init__()
        self.factor=factor

    def output_protocol(self, wire):
        super().output_protocol(wire)
        wire.CHANNELS=self.input_wire.CHANNELS
        wire.FORMAT=self.input_wire.FORMAT
        wire.RATE=self.input_wire.RATE/self.factor

    def process(self, data):
        data["data"]=numpy.max(data["data"].reshape(-1, self.factor),axis=1)
        return data

class Normalize(ThreadService, SourceService, ReceivingService):
    input_wire_class=FloatWire
    output_wire_class=FloatWire

    def __init__(self, baseline=0., scale=1.):
        super().__init__()
        self.scale=scale
        self.baseline=baseline

    def output_protocol(self, wire):
        super().output_protocol(wire)
        wire.CHANNELS=self.input_wire.CHANNELS
        wire.RATE=self.input_wire.RATE
        wire.FORMAT="float32"

    def process(self, data):
        data["data"]=numpy.clip(self.scale*(data["data"]-self.baseline), -1.,1., dtype="float32")
        return data

class Float2Int16(ThreadService, SourceService, ReceivingService):
    input_wire_class=FloatWire
    output_wire_class=Int16Wire

    def output_protocol(self, wire):
        super().output_protocol(wire)
        wire.CHANNELS=self.input_wire.CHANNELS
        wire.RATE=self.input_wire.RATE
        wire.FORMAT="int16"

    def process(self, data):
        data["data"]=numpy.array(32767*data["data"]).astype("int16")
        return data
          
class Int162Raw(ThreadService, SourceService, ReceivingService):
    input_wire_class=Int16Wire
    output_wire_class=RawWire

    def output_protocol(self, wire):
        super().output_protocol(wire)
        wire.CHANNELS=self.input_wire.CHANNELS
        wire.RATE=self.input_wire.RATE
        wire.FORMAT="int16"

    def process(self, data):
          data["data"]=numpy.array(data["data"]).tobytes()
          return data

class Int162Float(ThreadService, SourceService, ReceivingService):
    input_wire_class=Int16Wire
    output_wire_class=FloatWire

    def output_protocol(self, wire):
        super().output_protocol(wire)
        wire.CHANNELS=self.input_wire.CHANNELS
        wire.RATE=self.input_wire.RATE
        wire.FORMAT="float32"

    def process(self, data):
        data["data"]=data["data"].astype("float32")/32768
        return data




#!/usr/bin/env python
"""
gammalapp is a utility for acquisition of gamma spectrograph data and common analysis tasks. 
"""
import os
import argparse

from datetime import datetime

try:
  import matplotlib
  matplotlib.use('TkAgg')
except:
  pass

from gammalab import main
from gammalab.util import read_calibration_file
from gammalab.acquisition import SoundCard, FileReplay
from gammalab.transform import DownSampleMaxed
from gammalab.transform import Normalize, SecondOrder
from gammalab.analysis import PulseDetection, FittedPulseDetection
from gammalab.analysis import AggregateHistogram
from gammalab.analysis import Count, DoseCount
from gammalab.backend import PlotHistogram, CountPlot, PulsePlot
from gammalab.backend import SaveRaw
from gammalab.backend import SoundCardPlay
from gammalab.backend import Monitor


def run(threshold=0.003, nchannels=500, vmax=2000., offset=0, scale=5000., 
        drift=0., runtime=None, outfile=None, log=True, do_plot=False,
        input_device_name="", inputfile=None, realtime=True,
        fitpulse=False, fit_threshold=0.95, raw_values=False,
        baseline=0., negative_peaks=False, amplitude=1., plot_count=False,
        histogram_mode="normal", background="", plot_pulses=False,
        time_normalized=False, plot_excess=False,sound_api="soundcard", dose=False, 
        detector_mass=None, list_devices=False, raw_output_file="", output_device_name="", 
        monitor=False, detect=True):

    if list_devices:
        print("Input devices:")
        for name, id_ in SoundCard(sound_api=sound_api).devices().items():
            print(f"  {name}: {id_}")
        print("Output devices:")
        for name, id_ in SoundCardPlay(sound_api=sound_api).devices().items():
            print(f"  {name}: {id_}")
        exit(0)

    detect=detect or do_plot or plot_pulses or plot_count or fitpulse or (outfile is not None) or dose or raw_values

    if dose and raw_values:
        raise Exception("Dose estimate cannot be done on raw values (remove --dose or --raw")

    if raw_values:
        scale=1.
        vmax=1.
    else:
        if os.path.isfile("gammalab.ini"):
            offset,scale,drift=read_calibration_file("gammalab.ini")
        
        print(f"using offset={offset:5.2f}, scale={scale:5.2f}, drift={drift:5.2f}")

    if inputfile is not None:
        source=FileReplay(filename=inputfile, realtime=realtime)
    else:
        source=SoundCard(sound_api=sound_api,input_device_name=input_device_name)

    if monitor:
        monitor_=Monitor()
        downsample=DownSampleMaxed(factor=8)
            
        source.plugs_into(downsample)
        downsample.plugs_into(monitor_)

    if detect:

        normalize=Normalize(baseline=baseline, scale=-amplitude if negative_peaks else amplitude)
    
        pulse_file=outfile+".pulses" if outfile is not None else None
        if fitpulse:
            detect=FittedPulseDetection(threshold=threshold, fit_threshold=fit_threshold,
                                        outfile=pulse_file,emit_pulse_shapes=True)
        else:
            detect=PulseDetection(threshold=threshold, 
                                  outfile=pulse_file,emit_pulse_shapes=True)
        
        histogram=AggregateHistogram(nchannels=int(nchannels),
                                     vmin=threshold*scale, 
                                     vmax=vmax, 
                                     outfile=outfile+".histogram" if outfile is not None else None,
                                     histogram_mode=histogram_mode)
        
        source.plugs_into(normalize)
        normalize.plugs_into(detect)
            
        if raw_values:
            detect.plugs_into(histogram)
            detect_=detect
        else:
            calibrate=SecondOrder(offset=offset, scale=scale, drift=drift)
            detect.plugs_into(calibrate)
            calibrate.plugs_into(histogram)
            detect_=calibrate
    
        if dose:
            count=DoseCount(outfile=outfile+".dosecounts" if outfile is not None else None, runtime=runtime, detector_mass=detector_mass)
            detect_.plugs_into(count)
        else:
            count=Count(outfile=outfile+".counts" if outfile is not None else None, runtime=runtime)
            detect_.plugs_into(count)

    if plot_pulses:
        pulseplot=PulsePlot(nplot=5,
                            interval=1000)
        detect.plugs_into(pulseplot)

    if plot_count:
        countplot=CountPlot(outfile=outfile+".count" if outfile is not None else None)
        count.plugs_into(countplot)

    if do_plot:
        plothistogram=PlotHistogram(xmin=0, 
                                    xmax=vmax, 
                                    outfile=outfile,
                                    log=log,
                                    background=background,
                                    time_normalized=time_normalized,
                                    plot_excess=plot_excess)
        histogram.plugs_into(plothistogram)

    if raw_output_file!="":
        output_file=raw_output_file+"."+datetime.utcnow().strftime("%Y%m%d%H%M%S")
        save=SaveRaw(output_file)
        source.plugs_into(save)

    if output_device_name!="":
        playback=SoundCardPlay(sound_api=sound_api,output_device_name=output_device_name)
        source.plugs_into(playback)

    main(runtime)

def new_argument_parser():
    "Parse command line arguments"
    parser = argparse.ArgumentParser( formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                      description=__doc__, epilog="Note: if a gammalab.ini file is present and --raw not given, the calibration data in the file is used to determine offset,scale and drift")

    parser.add_argument(
        '--runtime',
        dest='runtime',
        default=None,
        type=float,
        help='runtime in seconds',
    )
    parser.add_argument(
        '--input-device',
        dest='input_device_name',
        default="",
        type=str,
        help='select input device by (fuzzy matched) name',
    )
    parser.add_argument(
        '--output-device',
        dest='output_device_name',
        default="",
        type=str,
        help='optional output device for playback (fuzzy matched by name)',
    )
    parser.add_argument(
        '--input-file',
        dest='inputfile',
        default=None,
        help='(optional) analyze from input file',
    )
    parser.add_argument(
        '--plot',
        dest='do_plot',
        action="store_true",
        help='show histogram plot',
    )
    parser.add_argument(
        '--plot-count',
        dest='plot_count',
        action="store_true",
        help='show plot of counts',
    )
    parser.add_argument(
        '--plot-pulses',
        dest='plot_pulses',
        action="store_true",
        help='Plot sampling of pulse shapes',
    )    
    parser.add_argument(
        '--log',
        dest='log',
        action="store_true",
        help='plot histogram with logarithmic y-axis',
    )
    parser.add_argument(
        '--raw',
        dest='raw_values',
        action="store_true",
        help='histogram of normalized raw detector values (no calibration)',
    )
    parser.add_argument(
        '--nchannels',
        dest='nchannels',
        default=200,
        type=int,
        help='number of channels (bins) in the histogram',
    )
    parser.add_argument(
        '--max_energy',
        dest='vmax',
        default=2000,
        type=float,
        help='maximum energy bin for histogram (keV, ignored for --raw)',
    )
    parser.add_argument(
        '--baseline',
        dest='baseline',
        default=0.,
        type=float,
        help='signal baseline (raw value)',
    )
    parser.add_argument(
        '--negative_peaks',
        dest='negative_peaks',
        action="store_true",
        default=False,
        help='inverted (negative peak) signal',
    )
    parser.add_argument(
        '--amplitude',
        dest='amplitude',
        default=1.,
        type=float,
        help='max signal amplitude for normalization (raw value)',
    )
    parser.add_argument(
        '--threshold',
        dest='threshold',
        default=0.003,
        type=float,
        help='(noise) threshold for pulse detection (normalized raw value)',
    )              
    parser.add_argument(
        '--offset',
        dest='offset',
        default=0.,
        type=float,
        help='signal energy offset (in keV)',
    )
    parser.add_argument(
        '--scale',
        dest='scale',
        default=5400,
        type=float,
        help='signal energy scale (in keV, ignored for --raw)',
    )
    parser.add_argument(
        '--drift',
        dest='drift',
        default=0,
        type=float,
        help='second order correction parameter (fractional deviation at raw value 1)',
    )
    parser.add_argument(
        '--fitpulse',
        dest='fitpulse',
        action="store_true",
        help='use pulse fitting for amplitude?',
    )       
    parser.add_argument(
        '--fit-threshold',
        dest='fit_threshold',
        default=0.95,
        type=float,
        help='minimum peak to determine amplitude through fitting (normalized raw value)',
    )    
    parser.add_argument(
        '--output-file',
        dest='outfile',
        default=None,
        help='output file root for analysis results (histogram etc)',
    )    
    parser.add_argument(
        '--histogram-mode',
        dest='histogram_mode',
        default="normal",
        type=str,
        help='mode for histogram binning (normal, proportional, semiprop, quadratic)',
    )
    parser.add_argument(
        '--time-normalized',
        dest='time_normalized',
        action="store_true",
        help='plot time normalized histogram',
    )
    parser.add_argument(
        '--plot-excess',
        dest='plot_excess',
        action="store_true",
        help='Plot excess counts in histogram (needs --background)',
    )        
    parser.add_argument(
        '--background',
        dest='background',
        default="",
        type=str,
        help='optional filename for background spectrum in histogram plot (recommended to add --time_normalized)',
    )     
    parser.add_argument(
        '--realtime',
        dest='realtime',
        action="store_true",
        help='file input in realtime',
    )       
    parser.add_argument(
        '--sound-api',
        dest='sound_api',
        default="SoundCard",
        help='Sound Card API to use [soundcard, sounddevice]',
    )    
    parser.add_argument(
        '--dose',
        dest='dose',
        action="store_true",
        help='show dose estimate',
    )       
    parser.add_argument(
        '--detector-mass',
        dest='detector_mass',
        default=None,
        type=float,
        help='detector mass (only needed for dose estimate) [0.01] ',
    )
    parser.add_argument(
        '--raw-output-file',
        dest='raw_output_file',
        default="",
        type=str,
        help='optionally record raw data stream to provided filename (for later analysis) ',
    )
    parser.add_argument(
        '--list-devices',
        dest='list_devices',
        action="store_true",
        default=False,
        help='list available input and output devices',
    )
    parser.add_argument(
        '--monitor',
        dest='monitor',
        action="store_true",
        default=False,
        help='monitor raw detector output',
    )
    return parser

if __name__=="__main__":
    args = new_argument_parser().parse_args()
    run(**vars(args))




from .raw import SaveRaw
from .wav import SaveWav
from ._soundcard import SoundCardPlay as _SoundCardPlay
from ._sounddevice import SoundDevicePlay
from .plot import Monitor, PlotHistogram, CountPlot, PulsePlot

def SoundCardPlay(sound_api="soundcard", **kwargs):
    if sound_api.lower()=="soundcard":
        return _SoundCardPlay(**kwargs)
    elif sound_api.lower()=="sounddevice":
        return SoundDevicePlay(**kwargs)
    else:
        raise Exception(f"Unknown sound API {sound_api}")



from ..service import SourceService, ThreadService
from ..wire import FloatWire

import numpy
import time
from contextlib import contextmanager
import signal

@contextmanager
def timeout(duration, message):
    def timeout_handler(signum, frame):
        raise TimeoutError(message)
    signal.signal(signal.SIGALRM, timeout_handler)
    signal.alarm(duration)
    try:
        yield
    finally:
        signal.alarm(0)


class SoundDevice(ThreadService, SourceService):
    output_wire_class=FloatWire

    @staticmethod
    def devices():
        """ return dict with names and ids of sound devices for acquisition (aka microphones) """
        try:
            import sounddevice
        except Exception as ex:
            raise Exception( f"import error: {str(ex)}")
        result=dict()
        for m in sounddevice.query_devices():
          if m["max_input_channels"]>0:
              result[m["name"]]=m["index"]
        return result
  
    def __init__(self, frames_per_buffer=2048, input_device_index=None, input_device_name="",
                 sample_rate=48000, sample_format="float32"):
        self.CHANNELS=1
        self.RATE=sample_rate
        if sample_format != "float32":
          raise Exception("SoundCard only supports float32 sample_format")
        self.FORMAT=sample_format
        self.frames_per_buffer=frames_per_buffer
        self.input_device_index=input_device_index
        self.input_device_name=input_device_name
        super().__init__()

    def start_process(self):
        global sounddevice
        try:
            import sounddevice
        except Exception as ex:
            self.print_message( f"import error: {str(ex)}")
        super().start_process()

    def _process(self):
        stream=sounddevice.InputStream(
          samplerate=self.RATE, blocksize=self.frames_per_buffer, channels=self.CHANNELS,
          device=self.input_device_index or self.input_device_name,
          dtype="float32"
          )
        name=sounddevice.query_devices(stream.device)["name"]
        self.print_message( f"opening {name} for data acquisition")
        t0=time.time()
        total_samples=0
        stream.start()
        buffer=numpy.zeros(self.frames_per_buffer,dtype=numpy.float32)
        while not self.stopped:    
            try:
                _name=sounddevice.query_devices(name)["name"] # this can raise exception if device dissapears
                assert _name==name # double check
            except:
                raise Exception("Soundcard acquisition device possibly disconnected")

            buffer[:]=0
            try:
                frames_read=0
                with timeout(1,"timeout on data acquisition, possible soundcard disconnect"):
                    while frames_read<self.frames_per_buffer:
                      available=stream.read_available
                      request=min(available, self.frames_per_buffer-frames_read)
                      if request>0:
                          _data, overflowed=stream.read(request)
                          buffer[frames_read:frames_read+request]=_data[:,0] # flatten because we have hardcoded nchannels=1
                          frames_read+=request
                      time.sleep(self.frames_per_buffer/self.RATE/2)
                
                t_wall=time.time()-t0
                if overflowed:
                    self.print_message("data acquisition overflow")
            except Exception as ex:
                self.print_message( f"error: {str(ex)}")
                self.stopped=True
            if frames_read==0:
                self.print_message("no data")
                self.stopped=True
            else:
                data=dict(data=buffer.copy(), wallclock_time=t_wall)

            if (not self.stopped and
                data is not None):
                self.send_output(data)
                total_samples+=len(data["data"])

        self.send_output(None)
        t_wall=time.time()-t0
        t_sample=total_samples/self.RATE
        self.print_message(f"wallclock(s): {t_wall:6.2f}, sample time(s): {t_sample:6.2f},  difference: {100*(t_wall-t_sample)/t_wall:4.2f}%")

        stream.stop()
        self.stopped=True

    def output_protocol(self, wire):
        super().output_protocol(wire)
        wire.CHANNELS=self.CHANNELS
        wire.RATE=self.RATE
        wire.FORMAT=self.FORMAT



from ..service import ThreadService, SourceService, ReceivingService
from ..wire import PulseWire, HistogramWire

import numpy
import pickle

class AggregateHistogram(ThreadService,ReceivingService, SourceService):
    input_wire_class=PulseWire
    output_wire_class=HistogramWire
    def __init__(self, nchannels=100, vmin=0.003, vmax=1., outfile="histogram",
        histogram_mode="normal"):
        super().__init__()

        self.vmin=vmin
        self.vmax=vmax
        self.nchannels=nchannels
        self.outfile=outfile
        self.histogram_mode=histogram_mode
        self.bins=None

    def get_proportional_bins(self):
        if self.vmin==0:
            raise Exception("For proportional bins, vmin cant be zero")
        fac=(self.vmax/self.vmin)**(1./self.nchannels)        
        bins=[]
        x=self.vmin
        for i in range(self.nchannels):
            bins.append(x)
            x=x*fac
        bins.append(self.vmax)
        
        return numpy.array(bins)

    def get_semi_proportional_bins(self):
        nextra=int(0.333*self.nchannels)
        if self.vmin==0:
            raise Exception("For proportional bins, vmin cant be zero")
        
        nchannels2=self.nchannels+nextra
        
        fac=(self.vmax/self.vmin)**(1./(nchannels2))        
        bins=[]
        x=self.vmin
        for i in range(nchannels2):
            bins.append(x)
            x=x*fac
        bins.append(self.vmax)
        bins=bins[nextra:]
        bins=numpy.array(bins)

        bins=bins/bins[0]*self.vmin
        l=bins[-1]-bins[0]
        bins=bins[0]+(bins-bins[0])/l*(self.vmax-self.vmin)

        return numpy.array(bins)

    def get_quadratic_bins(self):
        nn=2
        x=numpy.arange(self.nchannels+1)
        bins= self.vmax*( (x/self.nchannels)+(self.vmin/self.vmax)**(1./nn)) **nn / (1+(self.vmin/self.vmax)**(1/nn))**nn
        return bins
  
    def get_histogram(self, data):
        if self.bins is not None:
            return numpy.histogram(data, bins=self.bins)
        if self.histogram_mode=="normal":
            return numpy.histogram(data, bins=self.nchannels, 
                range=(self.vmin,self.vmax))
        if self.histogram_mode=="proportional":
            bins=self.get_proportional_bins()
            return numpy.histogram(data, bins=bins)
        if self.histogram_mode=="semiprop":
            bins=self.get_semi_proportional_bins()
            return numpy.histogram(data, bins=bins)
        if self.histogram_mode=="quadratic":
            bins=self.get_quadratic_bins()
            return numpy.histogram(data, bins=bins)
        raise Exception(f"Unknown histogram_mode {self.histogram_mode}")

    def start_process(self):
        self.hist, self.bins=self.get_histogram([])
        self.total_time=0
        super().start_process()

    def output_protocol(self, wire):
        super().output_protocol(wire)
        wire.nchannels=self.nchannels
        wire.vmin=self.vmin
        wire.vmax=self.vmax
        wire.unit=self.input_wire.unit
        wire.histogram_mode=self.histogram_mode

    def process(self, data):
        signal=[x[1] for x in data["pulses"]]
        hist, bins=self.get_histogram(signal)
        
        self.hist=self.hist+hist
        self.bins=bins
        self.total_time=data["total_time"]

        return self.outdata

    @property
    def outdata(self):
        return dict(hist=self.hist, bins=self.bins, total_time=self.total_time, 
                    unit=self.input_wire.unit)        

    def cleanup(self):
        if self.outfile is not None:
            outfile=self.outfile+".pkl"
            with open(outfile,"wb") as f:
                pickle.dump(self.outdata,f)
            self.print_message(f"Data written to {outfile}")
        super().cleanup()



from ..service import ReceivingService, SourceService, ThreadService
from ..wire import FloatWire, PulseWire

from functools import partial
import numpy
import pickle
    
class PulseDetection(ThreadService, SourceService, ReceivingService):
    input_wire_class=FloatWire
    output_wire_class=PulseWire
    def __init__(self, threshold=0.005, window=24*1024, emit_pulse_shapes=False, outfile=None):
        super().__init__()
        self.threshold=threshold
        self.window=window
        self.emit_pulse_shapes=emit_pulse_shapes
        self._x=numpy.arange(window)
        self.outfile=outfile
        self.all_pulses=[]

    def output_protocol(self, wire):
        super().output_protocol(wire)
        wire._emit_pulse_shapes=self.emit_pulse_shapes
        wire.unit="raw value"

    def start_process(self):
        self.data=numpy.zeros(self.window, dtype=self.input_wire.FORMAT)
        self.ndata=0
        self.RATE=self.input_wire.RATE
        self.itime=0
        self.dtime=self.window/(1.*self.RATE)
        self.wallclock_time=0
        super().start_process()
  
    def amplitude_and_quality(self, start,end):
        return numpy.max(self.data[start:end]),0

    def detect_pulses(self):
        larger=self.data>=self.threshold
        smaller=self.data<self.threshold
        
        up=smaller[:-1]*larger[1:]
        down=larger[:-1]*smaller[1:]
        
        up=numpy.argwhere(up) 
        down=numpy.argwhere(down) 
        
        up=up[:,0]
        down=down[:,0]
                
        if larger[0]: # discard first down crossing edge
          down=down[1:]
        if larger[-1]: # discard last up crossing edge
          up=up[:-1]
        
        if len(up)!=len(down):
            raise Exception("at this point, number of up and down edges must be same")
       
        #~ assert numpy.all(up<down)
        #~ assert numpy.all(up[1:]>down[:-1])
        #~ print "number of pulses:", len(up)
        
        pulses=[]
        for start,end in zip(up,down):
          if end-start>5:
            time=start/(1.*self.RATE)+self.itime
            try:
              amplitude, sigma=self.amplitude_and_quality(start,end)
            except Exception as ex:
              self.print_message("Pulse detection error condition: %s"%str(ex))
            
            width=end-start

            if self.emit_pulse_shapes:
                pulse=self.data[max(start-5,0):end+15].copy()
                pulses.append((time,amplitude, width, sigma, pulse))
            else:
                pulses.append((time,amplitude, width, sigma))

        return pulses

    def process(self, data):
        self.wallclock_time=data["wallclock_time"]
        data=data["data"]
        detect_ran=False
        pulses=[]
        dtime=0.
        while len(data)>0:
            start=self.ndata
            end=min(self.window,start+len(data))
            self.data[start:end]=data[:end-start]
            data=data[end-start:]
            self.ndata=end

            # if detection buffer filled, run detection
            if self.ndata>=self.window:
                pulses.extend(self.detect_pulses())
                detect_ran=True

                #update itime and dtime
                self.itime+=self.dtime
                dtime+=self.dtime
                
                self.ndata=0

        if self.outfile is not None:
            self.all_pulses.extend(pulses)

        if detect_ran:
            result=dict(pulses=pulses, total_time=self.itime, dtime=dtime, rate=self.RATE, wallclock_time=self.wallclock_time)
        else:
            result=None # do not send data if detection did not run
       
        return result

    @property            
    def outdata(self):
        return dict(pulses=self.all_pulses, total_time=self.itime, rate=self.RATE, wallclock_time=self.wallclock_time)

    def cleanup(self):
        if self.outfile is not None:
            outfile=self.outfile+".pkl"
            with open(outfile,"wb") as f:
                pickle.dump(self.outdata,f)
            self.print_message(f"Data written to {outfile}")
        super().cleanup()


def pulse(x, A, x0,  y, tau):
        return y+(x>=x0)*A*(x-x0)**2*numpy.exp(-(x/tau))+(x<x0)*0.


class FittedPulseDetection(PulseDetection):
    """ 
    detects pulses and determines amplitude by fitting a predifined 
    pulse shape. x**2*exp(-x/tau).

    In principle there could be two reasons to use fitted pulse shapes: in 
    case the signal is undersampled the maximum can fall between two 
    samples or if the signal saturates the ADC and the maximum can still be 
    recovered by fitting the rising and falling edges. 

    """
    def __init__(self, threshold=0.005, window=24*1024, emit_pulse_shapes=False,outfile=None, 
                  fit_threshold=0., signal_noise=0.0035, pulse_decay_time=2.7247/48000):

        super().__init__(threshold=threshold, 
                                                   window=window, 
                                                   emit_pulse_shapes=emit_pulse_shapes,
                                                   outfile=outfile)

        self.print_message("currently assumes very specific pulse shape, x**2 exp(-x/tau)")

        self.fit_threshold=fit_threshold
        self.sigma=signal_noise
        self.pulse_decay_time=pulse_decay_time

    def _fit_pulse(self, x, yp):
        global scipy, curve_fit
        
        A_=numpy.max(yp) # initial amplitude
        x0_=4.5 # initial timeshift
        tau=self.pulse_decay_time*self.input_wire.RATE
        
        template=partial(pulse, y=0., tau=tau) # fix base level (=0) and decay time
        
        a=(yp<0.95) * (yp>A_/5.) # select range
        x_=x[a]
        yp_=yp[a]
        
        sigma_=self.sigma*numpy.ones_like(x_) # weights
      
        (A,x0),cov=curve_fit(template, x_,yp_, p0=(A_,x0_), sigma=sigma_)
          
        AA=A*numpy.exp(-x0/tau)
        e=4*AA*tau**2*numpy.exp(-2.) # peak value
    
        p_=template(x_,A,x0)
        sigma_=numpy.sum((p_-yp_)**2)/len(yp_)
        sigma_=sigma_**0.5 / self.sigma # normalized mean error
    
        return e, sigma_
  
    def amplitude_and_quality(self, start,end):
        data=self.data[max(start-5,0):end+15]
        max_data=numpy.max(self.data[start:end])
        if max_data>self.fit_threshold:
            try:
                return self._fit_pulse(self._x[:len(data)], data)
            except Exception as ex:
                self.print_message("Fitted pulse detection error: %s"%str(ex))
                return max_data,0.
        else:
            return max_data,0.

    def start_process(self):
        global scipy, curve_fit
        
        try:
            import scipy
            from scipy.optimize import curve_fit
        except Exception as ex:
            self.print_message( f"import error: {str(ex)}")
            #~ self.stopped=True
            #~ self.done=True

        super().start_process()



from ..service import ThreadService, SourceService
from ..wire import FloatWire, Int16Wire

import time
import numpy
import wave

format_nbytes=dict(int16=2, float32=4)

class RawReplay(ThreadService, SourceService):
    output_wire_class=FloatWire
    def __init__(self, filename="data.raw", frames_per_buffer=1024, realtime=True, 
        sample_rate=48000, sample_format="float32"):
        super().__init__()
        self.CHANNELS=1
        self.RATE=sample_rate
        assert sample_format=="float32"
        self.FORMAT=sample_format
        self.filename=filename
        self.frames_per_buffer=frames_per_buffer
        self.realtime=realtime
        self.total_frames=0

    def output_protocol(self, wire):
        super().output_protocol(wire)      
        wire.CHANNELS=self.CHANNELS
        wire.RATE=self.RATE
        wire.FORMAT=self.FORMAT
    
    def readframes(self,n):
        return self._file.read(format_nbytes[self.FORMAT]*n)
    
    def process(self, data=None):
        if self.realtime:
          time.sleep(self.frames_per_buffer/(1.*self.RATE))
          t=time.time()
          dt=t-self.t0
          self.t0=t
          n=int(dt*self.RATE/16)*16
        else:
          n=int(self.frames_per_buffer/16)*16
        data=self.readframes(n)
        self.total_frames+=n
        if len(data)==0:
          self.print_message("End of file")
          self.stopped=True
          return None
        return dict(data=numpy.frombuffer(data, dtype=self.FORMAT), wallclock_time=self.total_frames/self.RATE)
        
    def start_process(self):
        self.t0=time.time()
        self.print_message(f"reading from {self.filename}")
        with open(self.filename, "rb") as self._file:
            super().start_process()


format_from_width={2 : "int16", 4 : "float32"}

class WavReplay(RawReplay):
    output_wire_class=Int16Wire
    def __init__(self, filename, frames_per_buffer=1024, realtime=True):
        super().__init__()
        self.filename=filename
        self.frames_per_buffer=frames_per_buffer
        self.realtime=realtime
        self._file=wave.open(self.filename, "rb")
        self.CHANNELS=1
        assert self._file.getnchannels()==1
        self.RATE=self._file.getframerate()
        self.SAMPLEWIDTH=self._file.getsampwidth()
        self.FORMAT=format_from_width[self.SAMPLEWIDTH]
        assert self.FORMAT=="int16"
        
    def readframes(self,n):
        return self._file.readframes(n)

    def start_process(self):
        self.t0=time.time()
        self.print_message(f"reading from {self.filename}")
        self._process()
        
    def cleanup(self):
        self._file.close()
        super().cleanup()

def FileReplay(filename, **kwargs):
    if filename.endswith("wav"):
        return WavReplay(filename, **kwargs)
    return RawReplay(filename, **kwargs)



from ..service import ThreadService, ReceivingService
from ..wire import FloatWire, HistogramWire, CountWire, PulseWire

import numpy
import time
import random
import pickle

class _Plot(ThreadService, ReceivingService):
    def __init__(self, outfile=None, interval=250):                
        super().__init__()
        self.thread.daemon=True
        self.outfile=outfile
        self.interval=interval
        self.blit=False
        self.do_update=True

    def on_click(self,event):
        if event.button is MouseButton.RIGHT:
            self.do_update=not self.do_update

    def setup_plot(self):
        raise Exception("setup_plot not implemented")

    def update_plot(self, nframe, data=None):
        raise Exception("update_plot not implemented")

    def get_data(self, nframe, data=None):
        raise Exception("get_data not implemented")

    def _update_plot(self, nframe):            
        self.fig.canvas.flush_events()

        if self.done:
            return []
      
        if self.stopped:
            self.do_update=False
            if self.outfile is not None:
                outfile=self.outfile+'.png'
                self.fig.savefig(outfile)
                self.print_message(f"Image written to {outfile}")
            self.done=True
            return []

        data=self.get_data()

        if self.do_update==False or data is None:
            return self.artists

        return self.update_plot(nframe, data)
        
    def start_process(self):
        global FuncAnimation, pyplot, MouseButton
        try:
            from matplotlib.backend_bases import MouseButton
            from matplotlib.animation import FuncAnimation
            from matplotlib import pyplot    
        except Exception as ex:
            self.print_message(f"import error: {str(ex)}")
            
        self.fig, self.ax, self.artists=self.setup_plot()

        pyplot.connect('button_press_event', self.on_click)

        
        ani = FuncAnimation(self.fig, self._update_plot, interval=self.interval, blit=self.blit,cache_frame_data=False)

        pyplot.show(block=True)

        if self.outfile is not None:
            outfile=self.outfile+'.png'
            self.fig.savefig(outfile)
            self.print_message(f"Image written to {outfile}")

        self.cleanup()
        
    def close(self):
        # deamon process
        self.stop()

        while not self.done:
            time.sleep(0.1)


class Monitor(_Plot):
    input_wire_class=FloatWire

    def __init__(self, window=5, vmin=-0.01, vmax=1.1, outfile=None):                
        super().__init__(outfile=outfile)
        self.window=window
        self.vmin=vmin
        self.vmax=vmax
        self.blit=True

    def get_data(self):
        data=[]
        while True:
            _data=self.receive_input(False)
            if _data is None:
                break
            data.append(_data["data"])

        if data:
            data=numpy.concatenate(data)
        else:
            data=None
        
        return data

    def update_plot(self,nframe, data):

        while len(data)>0:          
            n=min(len(data),len(self.plotdata)-self.nplot)
            
            self.plotdata[self.nplot:self.nplot+n]=data[:n]
            data=data[n:]
            self.nplot=(self.nplot+n) % len(self.plotdata)
                            
        self.artists[0].set_ydata(self.plotdata)
        return self.artists
  
    def setup_plot(self):
        self.plotdata=numpy.zeros(int(self.input_wire.RATE*self.window), dtype=self.input_wire.FORMAT)
        self.plotx=numpy.arange(int(self.input_wire.RATE*self.window))/float(self.input_wire.RATE)

        fig, ax = pyplot.subplots(figsize=(8,4))
        fig.canvas.manager.set_window_title("GammaLab Monitor")

        ax.cla()
        lines=ax.plot(self.plotx,self.plotdata)
        ax.set_ylim(self.vmin,self.vmax)
        ax.set_xlabel("time (s)")
        ax.set_ylabel("level")

        self.nplot=0
        
        return fig,ax,lines
        
class PlotHistogram(_Plot):
    input_wire_class=HistogramWire
  
    def __init__(self, xmin=0, xmax=1., log=True, 
                    error_bars=True, outfile="histogram",
                    background="", yrange=1.e4, time_normalized=False,
                    plot_excess=False):        
        super().__init__(outfile=outfile)
        self.xmin=xmin
        self.xmax=xmax
        self.log=log
        self.error_bars=error_bars
        self.background=background #"background.histogram.pkl"
        self.yrange=yrange
        self.blit=True
        self.time_normalized=time_normalized
        self.plot_excess=plot_excess
        if plot_excess and background=="":
            raise Exception("plot excess needs background spectrum")

    def get_data(self):        
        data=None
        while True:
            _data=self.receive_input(False)
            if _data is None:
                break
            data=_data
 
        return data

    def convert_hist_bins(self,hist,bins,total_time):
        yerr=hist**0.5
        yerr=numpy.maximum(yerr,1)
        if self.time_normalized:
            hist=hist/total_time
            yerr=yerr/total_time
        if self.input_wire.histogram_mode!="normal":
            binsize=(bins[1:]-bins[:-1])
            hist=hist/(binsize)
            yerr=yerr/(binsize)
        return hist,bins,yerr

    def update_plot(self,nframe, data):

        hist=data["hist"]
        total_count=hist.sum()
        bins=data["bins"]
        total_time=data["total_time"]
        hist,bins,yerr=self.convert_hist_bins(hist,bins,total_time)
        if self.plot_excess:
            midpoints=(bins[1:]+bins[:-1])/2
            index=numpy.digitize(midpoints, self.bkgrnd_bins)
            index=numpy.maximum(index,1)
            index=numpy.minimum(index, len(self.bkgrnd_hist))
            bkgrnd_hist=self.bkgrnd_hist[index-1]
            hist=numpy.maximum(hist-bkgrnd_hist,0)

        self.bins=bins
                
        self._line.set_data(hist,bins)
        if self.error_bars:
            self._top.set_data(hist+yerr,bins)
            self._bot.set_data(hist-yerr,bins)

        if self.plot_excess:
            if self.input_wire.histogram_mode!="normal":
                binsize=(bins[1:]-bins[:-1])
                excess_counts=(hist*binsize).sum()
            else:
                excess_counts=hist.sum()              
            if self.time_normalized: excess_counts*=total_time
            self._text.set_text(f"time(s): {total_time:10.2f}\ntotal counts: {total_count:09d}\nexcess counts: {int(excess_counts):09d}")
        else:
            self._text.set_text(f"time(s): {total_time:10.2f}\ntotal counts: {total_count:09d}")

        if numpy.any(hist!=0): self._update_ylim(hist[hist!=0].min(), hist.max())

        return self.artists

    def _update_ylim(self, ymin, ymax, ax=None):
        if ax is None:
            ax=self.ax
        _ymin,_ymax=ax.get_ylim()
        ymin_=_ymin
        ymax_=_ymax
        
        if self.log:
            ymin_=10**(numpy.floor(numpy.log10(ymin))-0.2)
            ymax_=10**numpy.ceil(numpy.log10(ymax*1.2))
        else:
            ymin_=0
            ymax_=2**numpy.ceil(numpy.log2(ymax*1.2))
            
        if _ymax!=ymax_ or _ymin!=ymin_:
            ax.set_ylim(ymin_,ymax_)
            pyplot.draw()


    def background_hist_bins(self):
        with open(self.background,'rb') as f:
            data=pickle.load(f)
        hist=data["hist"]
        bins=data["bins"]
        total_time=data["total_time"]
        if self.input_wire.unit != data["unit"]:
            message=f"Background spectrum wrong unit, expect {self.input_wire.unit}, has {data['unit']}"
            raise Exception(message)
        if (len(bins)!=len(self.bins) or numpy.any(bins-self.bins>1.e-7)) and self.input_wire.histogram_mode=="normal": 
            message=f"Background spectrum binning does not match"
            raise Exception(message)

        return self.convert_hist_bins(hist,bins,total_time)
        

    def setup_plot(self):
        fig, ax = pyplot.subplots()
        fig.canvas.manager.set_window_title("GammaLab Histogram")
        ax.cla()
        
        nchannels=self.input_wire.nchannels
        vmin=self.input_wire.vmin
        vmax=self.input_wire.vmax
        hist,bins=numpy.histogram([], bins=nchannels, range=(vmin,vmax))
        self.bins=bins

        if self.error_bars:
            yerr=hist**0.5
            self._top=ax.stairs(numpy.maximum(hist+yerr,1),bins,ls=":",color="tab:orange", zorder=0)
            self._bot=ax.stairs(hist-yerr,bins,ls=":",color="tab:orange",zorder=5)
        self._line=ax.stairs(hist,bins,lw=2, zorder=10)
        if self.background:
            hist,bins,err=self.background_hist_bins()
            if self.plot_excess:
                self.bkgrnd_hist=hist
                self.bkgrnd_bins=bins
            else:
                self._bkgrnd_line=ax.stairs(hist,bins,lw=2., zorder=11, color="tab:grey", ls="--")

        if self.log:
            ax.set_yscale("log", nonpositive="clip")

        ylabel="counts"
        if self.plot_excess:
          ylabel="(excess) "+ylabel
        if self.time_normalized:
          ylabel=ylabel+" / s"

        if self.input_wire.histogram_mode=="normal":
            ylabel=ylabel+" / bin"
        else:
            ylabel=ylabel+f" / {self.input_wire.unit}"
        ax.set_ylabel(ylabel)

        self._text=ax.text(0.95,0.95,f"time(s): {0.:10.2f}\ntotal counts: {0:09d}", va='top', ha='right',transform=ax.transAxes)
            

        ax.set_xlabel(f"energy ({self.input_wire.unit})")
        ax.set_xlim(self.xmin,self.xmax)
        self.ymin, self.ymax=ax.get_ylim()
        if self.log:
            self.ymin=self.ymax/self.yrange
            ax.set_ylim(self.ymin,self.ymax)
        
        pyplot.tight_layout()
        
        if self.error_bars:
            artists=[self._text,self._line, self._top, self._bot]
        else:
            artists=[self._text,self._line]
        
        if self.background and not self.plot_excess:
            artists+=[self._bkgrnd_line]
        
        return fig,ax,artists
      
class CountPlot(_Plot):
    input_wire_class=CountWire

    def __init__(self, outfile=None):                
        super().__init__(outfile=outfile)
        self.time=[]
        self.count=[]
        self.total_time=0
        self.avgcps=0
        self.blit=True

    def get_data(self):        
        data=None
        while True:
            _data=self.receive_input(False)
            if _data is None:
                break
            data=_data
 
        return data
        

    def update_plot(self,nframe, data):
            
        cps=data["count_per_sec"]
        tbins=data["time_bins"]
        self.total_time=data["total_time"]
        interval=data["interval"]
              
        self.count=cps
        self.time=(tbins[1:]+tbins[0:-1])/2
        if self.total_time!=0:
            self.avgcps=self.count.sum()*interval/self.total_time
        else:
            self.avgcps=0.
        
        self.artists[0].set_data(self.time, self.count)
        self.artists[1].set_data([self.time[0],self.total_time], [self.avgcps,self.avgcps])

        xmin,xmax=self.ax.get_xlim()
        ymin,ymax=self.ax.get_ylim()
        _xmax=self.time[-1]
        _ymax=1.5*max(self.count)
        if _xmax>xmax or _ymax>ymax:   
            self.ax.set_xlim(0,_xmax)
            self.ax.set_ylim(0,_ymax)
            pyplot.draw()

        return self.artists
  
    def setup_plot(self):
        fig, ax = pyplot.subplots(figsize=(6,3))
        fig.canvas.manager.set_window_title("GammaLab Counts")

        ax.cla()
        artist1,=ax.plot(self.time,self.count, ".")
        artist2,=ax.plot([0,10],[self.avgcps,self.avgcps], "g--")
        ax.set_xlabel("time (s)")
        ax.set_ylabel("counts per sec")
        ax.set_xlim(0,10)
        ax.set_ylim(0,30)
        pyplot.tight_layout()

        return fig,ax,[artist1, artist2]


class PulsePlot(_Plot):
    input_wire_class=PulseWire

    def __init__(self, nplot=1, interval=250):                
        super().__init__(interval=interval)
        self.nplot=nplot

    def get_data(self):
        data=[]
        while True:
            _data=self.receive_input(False)
            if _data is None:
                break
            data.append(_data)

        if data:
            return data
        else:
            return None

    def update_plot(self,nframe, data):

        self.process(data)
            
        return self.update_plot_data()

    def process(self,data):      
      
        _data=None
        for d in data:
            if _data is None:
                _data=d
            else:
                _data["pulses"].extend(d["pulses"])
                assert _data["rate"]==d["rate"]
                if d["total_time"]>_data["total_time"]:
                    _data["total_time"]=d["total_time"]
                _data["dtime"]+=d["dtime"]
      
        pulses=_data["pulses"]
        
        if pulses and len(pulses[0])<5:
            self.print_message("PulseWire does not contain pulse shape")
        
        for i,p in enumerate(random.sample(pulses, k=min(self.nplot, len(pulses)))):
            self.data[i]=p[4]
  
    def update_plot_data(self):
        dmin=9999
        dmax=-9999
        tmin=9999
        tmax=-9999
        for d,line in zip(self.data, self.lines):
            time=(numpy.arange(0,len(d))-5)
            line.set_data(time,d)                  
            
            tmin=min(tmin,time.min())
            tmax=max(tmax,time.max())
            
            dmin=min(dmin,d.min())
            dmax=max(dmax,d.max())

        self.ax.set_xlim(tmin,tmax)
        self.ax.set_ylim(dmin-.15*(dmax-dmin),dmax+0.2*(dmax-dmin))

        return self.lines[0:len(self.data)]
            
    def setup_plot(self):
        fig, ax = pyplot.subplots(figsize=(5,2))
        fig.canvas.manager.set_window_title("GammaLab Pulse Monitor")

        self.data=[numpy.array([0]*40)]*self.nplot

        self.lines=[]        
        for d in self.data:
            t=(numpy.arange(0,len(d))-5)
            line,=ax.plot(t,d,"+-")
            self.lines.append(line)

        ax.set_xlim(-5,35)
        ax.set_ylim(-.15,1.2)


        ax.set_xlabel("time")
        ax.set_ylabel("level")
        ax.set_xticks([],[])
        ax.set_yticks([],[])
        
        #~ ax.yaxis.set_label_coords(0.0,0.5)
        #~ ax.xaxis.set_label_coords(0.5,0.0)
        
        #~ ax.set_title("selected pulses")
        
        pyplot.tight_layout()
        
        artists=self.lines
        
        return fig,ax,artists



#!/usr/bin/env python
"""
gammalab-histogram makes a histogram of detected peak heights. 
"""
import os
import argparse

try:
  import matplotlib
  matplotlib.use('TkAgg')
except:
  pass

from gammalapp import run

def new_argument_parser():
    "Parse command line arguments"
    parser = argparse.ArgumentParser( formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                      description=__doc__, epilog="Note: if a gammalab.ini file is present and --raw not given, the calibration data in the file is used to determine offset,scale and drift")

    parser.add_argument(
        '--runtime',
        dest='runtime',
        default=None,
        type=float,
        help='runtime in seconds',
    )
    parser.add_argument(
        '--input-device',
        dest='input_device_name',
        default="",
        type=str,
        help='select input device by (fuzzy matched) name',
    )
    parser.add_argument(
        '--input-file',
        dest='inputfile',
        default=None,
        help='(optional) analyze from input file',
    )
    parser.add_argument(
        '--plot',
        dest='do_plot',
        action="store_true",
        help='show histogram plot',
    )
    parser.add_argument(
        '--plot-count',
        dest='plot_count',
        action="store_true",
        help='show plot of counts',
    )
    parser.add_argument(
        '--plot-pulses',
        dest='plot_pulses',
        action="store_true",
        help='Plot sampling of pulse shapes',
    )    
    parser.add_argument(
        '--log',
        dest='log',
        action="store_true",
        help='plot histogram with logarithmic y-axis',
    )
    parser.add_argument(
        '--raw',
        dest='raw_values',
        action="store_true",
        help='histogram of normalized raw detector values (no calibration)',
    )
    parser.add_argument(
        '--nchannels',
        dest='nchannels',
        default=200,
        type=int,
        help='number of channels (bins) in the histogram',
    )
    parser.add_argument(
        '--max_energy',
        dest='vmax',
        default=2000,
        type=float,
        help='maximum energy bin for histogram (keV, ignored for --raw)',
    )
    parser.add_argument(
        '--baseline',
        dest='baseline',
        default=0.,
        type=float,
        help='signal baseline (raw value)',
    )
    parser.add_argument(
        '--negative_peaks',
        dest='negative_peaks',
        action="store_true",
        default=False,
        help='inverted (negative peak) signal',
    )
    parser.add_argument(
        '--amplitude',
        dest='amplitude',
        default=1.,
        type=float,
        help='max signal amplitude for normalization (raw value)',
    )
    parser.add_argument(
        '--threshold',
        dest='threshold',
        default=0.003,
        type=float,
        help='(noise) threshold for pulse detection (normalized raw value)',
    )              
    parser.add_argument(
        '--offset',
        dest='offset',
        default=0.,
        type=float,
        help='signal energy offset (in keV)',
    )
    parser.add_argument(
        '--scale',
        dest='scale',
        default=5400,
        type=float,
        help='signal energy scale (in keV, ignored for --raw)',
    )
    parser.add_argument(
        '--drift',
        dest='drift',
        default=0,
        type=float,
        help='second order correction parameter (fractional deviation at raw value 1)',
    )
    parser.add_argument(
        '--fitpulse',
        dest='fitpulse',
        action="store_true",
        help='use pulse fitting for amplitude?',
    )       
    parser.add_argument(
        '--fit-threshold',
        dest='fit_threshold',
        default=0.95,
        type=float,
        help='minimum peak to determine amplitude through fitting (normalized raw value)',
    )    
    parser.add_argument(
        '--outfile',
        dest='outfile',
        default=None,
        help='output file root',
    )    
    parser.add_argument(
        '--histogram-mode',
        dest='histogram_mode',
        default="normal",
        type=str,
        help='mode for histogram binning (normal, proportional, semiprop, quadratic)',
    )
    parser.add_argument(
        '--time-normalized',
        dest='time_normalized',
        action="store_true",
        help='plot time normalized histogram',
    )
    parser.add_argument(
        '--plot-excess',
        dest='plot_excess',
        action="store_true",
        help='Plot excess counts in histogram (needs --background)',
    )        
    parser.add_argument(
        '--background',
        dest='background',
        default="",
        type=str,
        help='optional filename for background spectrum in histogram plot (recommended to add --time_normalized)',
    )     
    parser.add_argument(
        '--realtime',
        dest='realtime',
        action="store_true",
        help='file input in realtime',
    )       
    parser.add_argument(
        '--sound-api',
        dest='sound_api',
        default="SoundCard",
        help='Sound Card API to use [soundcard, sounddevice]',
    )    
    parser.add_argument(
        '--dose',
        dest='dose',
        action="store_true",
        help='show dose estimate',
    )       
    parser.add_argument(
        '--detector-mass',
        dest='detector_mass',
        default=0.01,
        type=float,
        help='detector mass (only needed for dose estimate) [0.01] ',
    )
    return parser

if __name__=="__main__":
    args = new_argument_parser().parse_args()
    run(**vars(args))




#!/usr/bin/env python
"""
gammalab-monitor plots raw detector (soundcard) values. 
"""
import argparse
from datetime import datetime

try:
  import matplotlib
  matplotlib.use('TkAgg')
except:
  pass

from gammalapp import run

def new_argument_parser():
    "Parse command line arguments"
    parser = argparse.ArgumentParser( formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                      description=__doc__)

    parser.add_argument(
        '--runtime',
        dest='runtime',
        default=None,
        type=float,
        help='runtime in seconds',
    )
    parser.add_argument(
        '--input-device',
        dest='input_device_name',
        default="",
        type=str,
        help='select input device by (fuzzy matched) name',
    )
    parser.add_argument(
        '--output-device',
        dest='output_device_name',
        default="",
        type=str,
        help='optional output device for playback (fuzzy matched by name)',
    )
    parser.add_argument(
        '--list-devices',
        dest='list_devices',
        action="store_true",
        default=False,
        help='list available input and output devices',
    )
    parser.add_argument(
        '--output-file',
        dest='raw_output_file',
        default="",
        type=str,
        help='optionally record raw data stream to provided filename (for later analysis) ',
    )
    parser.add_argument(
        '--input-file',
        dest='inputfile',
        default=None,
        help='Optionally playback from input file',
    ) 
    parser.add_argument(
        '--sound-api',
        dest='sound_api',
        default="SoundCard",
        help='Sound Card API to use [soundcard, sounddevice]',
    )    
    return parser

if __name__=="__main__":
    args = new_argument_parser().parse_args()
    args=vars(args)
    args["monitor"]=True
    args["detect"]=False
    run(**args)



import sys
import pickle
import numpy

from ..service import ReceivingService, ThreadService, SourceService
from ..wire import PulseWire, CountWire
      
class DoseCount(ThreadService, ReceivingService, SourceService):
    input_wire_class=PulseWire
    output_wire_class=CountWire
    def __init__(self, outfile=None, runtime=None, interval=1., silent=False, detector_mass=None):
        super().__init__()
        self.outfile=outfile
        self.runtime=runtime
        self.tmax=2*interval
        self.tmin=0
        self.interval=interval
        self.total_count=0
        self.total_time=0
        self.recent_pulse_times=[] # all pulse times from last interval
        self.recent_pulse_energies=[] # all pulse energies of last interval
        self.silent=silent
        self.dose=0.
        self.detector_mass=detector_mass

    def connect_input(self, service):
        super().connect_input(service)
        assert self.input_wire.unit=="keV"
    
    @property
    def nbins(self):
        return int((self.tmax-self.tmin)/self.interval)
        
    def start_process(self):
        if self.detector_mass is None:
            self.print_message("detector mass not specified, assuming 0.01 kg")
            self.detector_mass=0.01
        self.factor=1000*1.6e-19/(self.detector_mass)*3600/1.e-6  # converts keV/s (calculated below) to J/kg/hr (=uSv/hr or rather uGy/hr)
        
        self.cps, self.tbins=numpy.histogram([], bins=self.nbins, 
            range=(self.tmin,self.tmax))
        self.dose_rate=numpy.histogram([], bins=self.nbins, 
            range=(self.tmin,self.tmax))[0]

        super().start_process() 
    
    def process(self, data):      

        self.total_time=data["total_time"]
        self.total_count+=len(data["pulses"])

        if self.tmax<self.total_time:
            self.tmax=self.tmax*2
        if self.runtime and self.tmax>self.runtime:
            self.tmax=self.runtime

        _cps=self.cps
        _dose_rate=self.dose_rate

        pulse_times=[d[0] for d in data["pulses"]]
        pulse_energies=[d[1] for d in data["pulses"]]
        self.recent_pulse_times.extend(pulse_times)
        self.recent_pulse_energies.extend(pulse_energies)
        self.recent_pulse_energies=[e for t,e in zip(self.recent_pulse_times, self.recent_pulse_energies) if t>self.total_time-self.interval]
        self.recent_pulse_times=[t for t in self.recent_pulse_times if t>self.total_time-self.interval]

        pulse_times=numpy.array(pulse_times)
        pulse_energies=numpy.array(pulse_energies)
        
        self.dose+=pulse_energies.sum()

        counts, self.tbins=numpy.histogram(pulse_times, bins=self.nbins, 
            range=(self.tmin,self.tmax))
        dose, dummy=numpy.histogram(pulse_times, bins=self.nbins, 
            range=(self.tmin,self.tmax), weights=pulse_energies)
        
        self.cps=counts/self.interval
        self.cps[0:len(_cps)]+=_cps

        self.dose_rate=dose/self.interval
        self.dose_rate[0:len(_dose_rate)]+=_dose_rate

        avgcps=self.total_count/(self.total_time)
        if self.total_time>0:
          cps=len(self.recent_pulse_times)/min(self.total_time, self.interval)
        else:
          cps=0

        avgdose_rate=self.factor*self.dose/self.total_time
        if self.total_time>0:
          dose_rate=self.factor*numpy.sum(self.recent_pulse_energies)/min(self.total_time, self.interval)
        else:
          dose_rate=0

        message=f"time: {self.total_time:.2f}, dose rate(uGy/hr): {dose_rate:5.2f}, average dr(uGy/hr): {avgdose_rate:5.2f}, total dose(uGy): {avgdose_rate*self.total_time/3600.:6.3f}"
        if not self.silent:
            self.print_message(message, end="\r")
        
        return self.outdata
    
    @property            
    def outdata(self):
        return dict(count_per_sec=self.cps, dose_rate=self.dose_rate, time_bins=self.tbins, total_time=self.total_time, interval=self.interval)
      
    def cleanup(self):
        if self.outfile is not None:
            outfile=self.outfile+".pkl"
            with open(outfile,"wb") as f:
                pickle.dump(self.outdata,f)
            self.print_message(f"Data written to {outfile}")
        super().cleanup()



import numpy
from ..service import ReceivingService, ThreadService
from ..wire import FloatWire

class SoundDevicePlay(ThreadService, ReceivingService):

    @staticmethod
    def devices():
        """ return dict with names and ids of sound devices for playback (aka speakers) """
        try:
            import sounddevice
        except Exception as ex:
            raise Exception( f"import error: {str(ex)}")
        result=dict()
        for m in sounddevice.query_devices():
          if m["max_output_channels"]>0:
              result[m["name"]]=m["index"]
        return result

    input_wire_class=FloatWire

    def __init__(self, frames_per_buffer=2048, output_device_index=None, output_device_name=""):
        self.frames_per_buffer=frames_per_buffer
        self.output_device_index=output_device_index
        self.output_device_name=output_device_name
        self.CHANNELS=1
        super().__init__()

    def start_process(self):
        global sounddevice
        try:
            import sounddevice
        except Exception as ex:
            self.print_message( f"import error: {str(ex)}")
        super().start_process()

    def _process(self):            
        stream=sounddevice.OutputStream(
          samplerate=self.input_wire.RATE, blocksize=self.frames_per_buffer, channels=self.CHANNELS,
          device=self.output_device_index or self.output_device_name,
          dtype="float32"
          )
        name=sounddevice.query_devices(stream.device)["name"]  
        self.print_message( f"opening {name} for audio output")
        stream.start()
        while not self.stopped:
            data=self.receive_input()
            if data is None:
                self.stopped=True
            else:
                data=data["data"]

            if not self.stopped:
                try:
                    underflowed=stream.write(data)
                except Exception as ex:
                    self.print_message( f"error: {str(ex)}")
                    self.stopped=True

        stream.stop()
        self.stopped=True




from ..service import SourceService, ThreadService
from ..wire import FloatWire

import time

class SoundCard(ThreadService, SourceService):
    output_wire_class=FloatWire

    @staticmethod
    def devices():
        """ return dict with names and ids of sound devices for acquisition (aka microphones) """
        try:
            import soundcard
        except Exception as ex:
            raise Exception( f"import error: {str(ex)}")
        result=dict()
        for m in soundcard.all_microphones():
          result[m.name]=m.id
        return result
  
    def __init__(self, frames_per_buffer=2048, input_device_index=None, input_device_name="",
                 sample_rate=48000, sample_format="float32"):
        self.CHANNELS=1
        self.RATE=sample_rate
        if sample_format != "float32":
          raise Exception("SoundCard only supports float32 sample_format")
        self.FORMAT=sample_format
        self.frames_per_buffer=frames_per_buffer
        self.input_device_index=input_device_index
        self.input_device_name=input_device_name
        super().__init__()

    def start_process(self):
        global soundcard
        try:
            import soundcard
        except Exception as ex:
            self.print_message( f"import error: {str(ex)}")
        super().start_process()

    def _process(self):
        mic=soundcard.get_microphone(self.input_device_index or self.input_device_name)
        self.print_message( f"opening {str(mic)} for data acquisition")
        name=mic.name
        t0=time.time()
        t_wall=0.
        total_samples=0
        with mic.recorder(samplerate=self.RATE, blocksize=self.frames_per_buffer, channels=self.CHANNELS) as recorder:
            while not self.stopped:    
                try:
                    _name=mic.name # this can raise exception if mic dissapears
                    assert _name==name # double check
                except:
                    raise Exception("Soundcard acquisition device possibly disconnected")

                try:
                    data=recorder.record(self.frames_per_buffer)
                    t_prev=t_wall
                    t_wall=time.time()-t0
                    if t_wall-t_prev>5:
                        self.print_message("warning: long pause in data acquisition (caused by e.g. process pause or entering sleep mode)")
                        self.print_message("(this can cause data corruption)")
                except Exception as ex:
                    self.print_message( f"error: {str(ex)}")
                    self.stopped=True
                if len(data)==0 or data is None:
                    self.print_message("no data")
                    self.stopped=True
                else:
                    data=dict(data=data.flatten(), wallclock_time=t_wall) # flatten because we have hardcoded nchannels=1
    
                if (not self.stopped and
                    data is not None):
                    self.send_output(data)
                    total_samples+=len(data["data"])
                
            self.send_output(None)
        t_wall=time.time()-t0
        t_sample=total_samples/self.RATE
        self.print_message(f"wallclock(s): {t_wall:6.2f}, sample time(s): {t_sample:6.2f},  difference: {100*(t_wall-t_sample)/t_wall:4.2f}%")

        self.stopped=True

    def output_protocol(self, wire):
        super().output_protocol(wire)
        wire.CHANNELS=self.CHANNELS
        wire.RATE=self.RATE
        wire.FORMAT=self.FORMAT



"""
  Calibration converts raw detector values to energies
"""

from ..service import ReceivingService, SourceService, ThreadService
from ..wire import PulseWire

import numpy

# todo: simplify with inheritance?

class Interpolate(ThreadService, SourceService, ReceivingService):
    input_wire_class=PulseWire
    output_wire_class=PulseWire

    def __init__(self, calibration, unit="keV"):
        super().__init__()
        self._xp=numpy.array([c[0] for c in calibration])
        self._yp=numpy.array([c[1] for c in calibration])
        try:
            assert(numpy.all(numpy.diff(self._xp) > 0))
        except AssertionError:
            raise Exception("provide ordered calibration dataset")
        self.unit=unit
        
    def output_protocol(self, wire):
        super().output_protocol(wire)
        wire.unit=self.unit

    def process(self, data):
        timestamps=[x[0] for x in data["pulses"]]
        signal=[x[1] for x in data["pulses"]]
        calibrated=numpy.interp(signal, self._xp, self._yp)
        data["pulses"]=list(zip(timestamps, calibrated))
        return data

class Scale(ThreadService, SourceService, ReceivingService):
    input_wire_class=PulseWire
    output_wire_class=PulseWire    

    def __init__(self, scale, unit="keV"):
        super().__init__()
        self.scale=scale
        self.unit=unit
        
    def output_protocol(self, wire):
        super().output_protocol(wire)
        wire.unit=self.unit

    def process(self, data):
        timestamps=[x[0] for x in data["pulses"]]
        scaled=[self.scale*x[1] for x in data["pulses"]]
        data["pulses"]=list(zip(timestamps, scaled))
        return data

class SecondOrder(ThreadService, SourceService, ReceivingService):
    input_wire_class=PulseWire
    output_wire_class=PulseWire    

    def __init__(self, scale, offset=0., drift=00, unit="keV"):
        super().__init__()
        self.input_wire=PulseWire()
        self.scale=scale
        self.drift=drift
        self.offset=offset
        self.unit=unit
                
    def output_protocol(self, wire):
        super().output_protocol(wire)
        wire.unit=self.unit

    def process(self, data):
        timestamps=[x[0] for x in data["pulses"]]
        scaled=[self.offset+self.scale*x[1]*(1+self.drift*x[1]) for x in data["pulses"]]
        data["pulses"]=list(zip(timestamps, scaled))
        return data



#!/usr/bin/env python
"""
gammalab-calibration-cal determines calibration coeff for a polynomial calibration curve. 
"""


import argparse
import numpy
from gammalab.util import get_calibration_coeff, calibration_pair

def new_argument_parser():

    parser = argparse.ArgumentParser(description='Determine scale and drift parameters from calibration data')
    parser.add_argument(
        dest='data', 
        default=None, 
        nargs="*", 
        type=calibration_pair,
        help='calibration data: a list of space sperated <raw value>,<energy> pairs, e.g.: 0.1,100 0.5,500.'
    )

    parser.add_argument(
        '--order',
        dest='order',
        default=2,
        type=int,
        help='order of calibration polynomial fit (1 or 2)',
    )
    parser.add_argument(
        '--no_offset',
        dest='no_offset',
        action="store_true",
        default=False,
        help='force offset to be 0.',
    )
    parser.add_argument(
        '--write_ini',
        dest='write_ini',
        action="store_true",
        default=False,
        help='write calibration data to a file (gammalab.ini)',
    )
  
    return parser

if __name__=="__main__":
    args=new_argument_parser().parse_args()
    offset,scale,drift=get_calibration_coeff(**vars(args))
    print("[gammalab-calibration-calc] Calibration parameters:")
    print(f" the offset is: {offset}")
    print(f" the scale is: {scale}")
    print(f" the drift is: {drift}")



from ._soundcard import SoundCard as _SoundCard
from ._sounddevice import SoundDevice
from .generator import Noise
from .file import RawReplay, WavReplay, FileReplay

def SoundCard(sound_api="soundcard", **kwargs):
    if sound_api.lower()=="soundcard":
        return _SoundCard(**kwargs)
    elif sound_api.lower()=="sounddevice":
        return SoundDevice(**kwargs)
    else:
        raise Exception(f"Unknown sound API {sound_api}")



from . import all_services, shared_output

import multiprocessing
import ctypes
import time

class ServiceError(Exception):
    pass

class Service:
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        all_services.add(self)
        self._output=shared_output
    
    def start(self):
        raise Exception("not implemented for %s"%str(self))
    def stop(self):
        raise Exception("not implemented for %s"%str(self))
    def close(self):
        raise Exception("not implemented for %s"%str(self))
    def plugs_into(self, other):
        raise ServiceError("service %s has no outputs"%str(self))
    def connect_to(self, other):
        raise ServiceError("service %s has no inputs"%str(self))
    def print_message(self, message, end=None):
        message=f'\33[2K[{self.__class__.__name__}] '+message
        message=dict(message=message,end=end)
        self._output.put(message)

class SourceService(Service):
    output_wire_class=None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.output_wires=[]
        if self.output_wire_class is None:
            raise Exception(f"SourceService {self.__class__.__name__} does not specify output_wire_class")

    def output_protocol(self, wire):
        assert isinstance(wire, self.output_wire_class)

    def connect(self, wire):
        self.output_protocol(wire)
        self.output_wires.append(wire)

    def send_output(self,data):
        for w in self.output_wires:
            try:
                w.put_nowait(data)
            except:
                self.print_message("%s buffer full"%str(self))

    def plugs_into(self, other):
        try:
          other.connect_input(self)
        except ServiceError as ex:
          raise ex      
        except AssertionError as ex:
          self.print_message(str(ex))
          raise Exception(f"Wiring fault! Trying to connect incompatible services {self.__class__.__name__} and {other.__class__.__name__}" )
        except AttributeError as ex:
          self.print_message(str(ex))
          raise Exception("Wiring fault! Wires connected out of order?")
        
class ReceivingService(Service):
    input_wire_class=None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if self.input_wire_class is None:
            raise Exception(f"ReceivingService {self.__class__.__name__} does not specify input_wire_class")
    def connect_input(self, service):
        self.input_wire=self.input_wire_class()
        service.connect(self.input_wire)
    def receive_input(self, block=True):
        if not hasattr(self,"input_wire"):
            self.print_message("Unconnected input")
            return None
        try:
            return self.input_wire.get(block)
        except multiprocessing.queues.Empty:
            return None
        except Exception as ex:
            self.print_message("receive_input error:")
            self.print_message(str(ex))
            return None

# multiple input/outputs will be implemented by connecting to/from special attributes:
# s1.output1.plugs_into(s2)

class ThreadService(Service):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.__stopped=multiprocessing.Value(ctypes.c_bool)
        self.__done=multiprocessing.Value(ctypes.c_bool)
        self.stopped=True
        self.done=False
        self.thread=multiprocessing.Process(target=self._start_process, 
                                            name=self.__class__.__name__)

    def start(self):
        self.stopped=False
        if not self.thread.is_alive():
            self.thread.start()

    def _start_process(self):
        try:
          self.start_process()
        except Exception as ex:
          message="Service process exception:"
          self.print_message(message)
          self.print_message(str(ex))
          if hasattr(self,"send_output"):
              self.send_output(None)
        self.cleanup()


    def start_process(self):
        self._process()

    def _process(self):
        while not self.stopped:
            if hasattr(self,"receive_input"):
                data=self.receive_input()
                if data is None:
                    self.stopped=True
            else:
                data=None

            # if stopped reveive input, but do not process, nor output

            if not self.stopped:
                data=self.process(data)

            if hasattr(self,"send_output"):
                if (not self.stopped and 
                    data is not None):  
                    self.send_output(data)
            
        if hasattr(self,"send_output"):
            self.send_output(None)
                    
        self.stopped=True
        
    def cleanup(self):
        self.stopped=True
        self.done=True

    def stop(self):
        self.stopped=True

    def close(self):
        self.stop()

        while not self.done:
            time.sleep(0.1)

        self.thread.terminate()
        self.thread.join()
        self.thread.close()

    @property
    def done(self):
        return self.__done.value

    @done.setter
    def done(self, value):
        self.__done.value=value

    @property
    def stopped(self):
        return self.__stopped.value

    @stopped.setter
    def stopped(self, value):
        self.__stopped.value=value




from multiprocessing import Queue as Queue

class Wire:
    _initialized=False
    _description="none"

    def __init__(self, **kwargs):
        self._queue=Queue(**kwargs)

# forward the necessary methods; note we are not deriving class from Queue since 
# multiprocessing Queue import is a factory method
        self.get=self._queue.get
        self.put=self._queue.put
        self.put_nowait=self._queue.put_nowait
        self.get_nowait=self._queue.get_nowait
        
        self.protocol=dict()
        self._initialized=True

    def __getstate__(self):
        d=self.__dict__.copy()
        d["_initialized"]=False
        return self.__dict__

    def __setstate__(self, d):
        for x in d:
          setattr(self, x, d[x])
        self._initialized=True

# attributes are forwarded to the protocol (after __init__)
    def __setattr__(self, name, value):
        if self._initialized:
            self.protocol[name] = value
        else:
            self.__dict__[name] = value

    def __getattr__(self, name):
        if name in self.protocol:
            return self.protocol[name]
        else:
            raise AttributeError(f"No attribute {name} in protocol of {self.__class__.__name__} ")

    def __delattr__(self, name):
        if name in self.protocol:
            del self.protocol[name]
        else:
            raise AttributeError(f"No attribute {name} in protocol of {self.__class__.__name__} ")

class RawWire(Wire):
    _description="Raw byte (string)"

class NumpyWire(Wire):
    _description="numpy array of flexible format (deprecated)"

class FloatWire(Wire):
    _description="numpy float array"

class Int16Wire(Wire):
    _description="numpy int16 array"

class PulseWire(Wire):
    _description="pulses"

class MessageWire(Wire):
    _description="output queue for terminal messages"

class HistogramWire(Wire):
    _description="dict with histogram"

class CountWire(Wire):
    _description="time, counts per sec"



from . import all_services, shared_output

import multiprocessing
import threading
import time

import os
if os.name == 'nt': # Only if we are running on Windows
    from ctypes import windll
    k = windll.kernel32
    k.SetConsoleMode(k.GetStdHandle(-11), 7) # fix ternminal

try:
    input=raw_input
except:
    pass

def output_thread():
    message=None
    while message!="[Shutdown] done":
        message=shared_output.get(True)
        end=None
        if isinstance(message, dict):
          end=message.get("end", None)
          message=message["message"]
        print(message, end=end)

def startup():
    print("[Startup] entry")

    t=threading.Thread(target=output_thread, name="output")
    t.daemon=True
    t.start()

    for s in all_services:
        s.start()

def shutdown():
    if all_services:
        shared_output.put("[Shutdown] shutting down")

    for s in all_services:
          s.stop()

    while all_services:
        s=all_services.pop()
        try:
          s.close()
        except Exception as ex:
          shared_output.put("[Shutdown] "+str(ex))

    shared_output.put("[Shutdown] done")

    time.sleep(0.5)

def check_user_input():
    input()

def main(timeout=None, debug=False):

    startup()

    if debug:
        shared_output.put("[Main] processes spawned:")
        for p in multiprocessing.active_children():
            shared_output.put(f"[Main] {str(p)}")

    shared_output.put("[Main] running")

    input_checker=threading.Thread(target=check_user_input)
    input_checker.daemon=True
    input_checker.start()
    input_checker.join(timeout)    
    
    shutdown()
    
    
